//
//  This file is part of SXML, a tool to automate the publication of
//  XML data through the Siena wide-area event notification system.
//
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Copyright (C) 2000-2001 University of Colorado
//
//  Author: Antonio Carzaniga <carzanig@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
// $Id: AttributeRule.java,v 1.3 2001/01/24 17:16:43 carzanig Exp $
//
package siena.sxml;

/** a rule to assign the value of a specific Siena attribute within an
 *  {@link EventMap}.
 *  
 *  The value of the Siena attribute in an {@link EventMap} is
 *  generated by a set of XML elements within an XML document (nodes
 *  is a document tree), selected by an XPath expression.  An
 *  <code>AttributeRule</code> defines the XPath expression, the type
 *  of the Siena attribute, and the method by which a single attribute
 *  value is derived from multiple selected nodes.
 *  
 *  <p>Example:
 *  <pre><code>
 *      EventMap map = new EventMap();
 *      AttributeRule rule = new AttributeRule();
 *
 *      rule.path = "/person/@name";
 *      rule.type = AttributeValue.BYTEARRAY;
 *      rule.multivalue = AttributeRule.SEPARATOR;
 *	rule.separator = ",";
 *
 *      map.putRule("names", rule);
 *  </code></pre>
 *  
 **/
public class AttributeRule {
    /** the value is derived from the first selected node **/ 
    public static final int	FIRST_ONLY	= 0;
    /** the value is derived from the last selected node **/ 
    public static final int	LAST_ONLY	= 1;
    /** the value is derived from all the selected CDATA nodes by
        concatenation **/
    public static final int	SEPARATOR	= 2;
    /** the value is derived from the XML representation of all the
        selected ELEMENT nodes **/
    public static final int	XML		= 3;

    /** selects the nodes that will produce the value of this Siena
     *  attribute.
     *  
     *  It is interpreted as an XPath.  Please, consult the <a
     *  href="http://www.w3.org/xpath/">XPath documentation</a> for
     *  additional information on syntax and semantics of XPath
     *  expressions.
     *  
     **/
    public String	path;

    /** Siena type identifier.
     *  
     *  valid identifiers are
     *  <code>siena.AttributeValue.BYTEARRAY</code>,
     *  <code>siena.AttributeValue.LONG</code>,
     *  <code>siena.AttributeValue.DOUBLE</code>, etc.  Please, 
     *  consult the <a href="http://www.cs.colorado.edu/serl/siena/software/java/index.html">Siena Java API documentation</a> for more information.
     **/ 
    public int		type;

    /** encoding of values from multiple selected nodes.
     *  
     *  valid encodings are:
     *  <ul>
     *
     *  <li><b>{@link AttributeRule#FIRST_ONLY}</b>: the value is that
     *      of the first element;
     *
     *  <li><b>{@link AttributeRule#LAST_ONLY}</b>: the value is that
     *      of the last element;
     *
     *  <li><b>{@link AttributeRule#SEPARATOR}</b>: the value is
     *      obtained by catenating the value of each selected node,
     *      with a given separator.  The given path must select
     *      <i>CDATA</i> nodes.
     *
     *  <li><b>{@link AttributeRule#XML}</b>: the value is obtained by
     *      catenating all the XML representation of each node.  The
     *      given * path must select <i>Element</i> nodes.  
     *
     * </ul>
     **/
    public int		multivalue;

    /** separator for encoding multiple values in a single string
     *  
     *  this value is ignored if <code>multivalue != SEPARATOR</code>.
     **/
    public String	separator;
};

