//
//  This file is part of SXML, a tool to automate the publication of
//  XML data through the Siena wide-area event notification system.
//
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Copyright (C) 2000-2001 University of Colorado
//
//  Author: Antonio Carzaniga <carzanig@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
// $Id: EventMap.java,v 1.4 2001/07/27 15:37:48 carzanig Exp $
//
package siena.sxml;

import java.util.*;
import java.io.*;
import javax.xml.transform.TransformerException;
import javax.xml.transform.OutputKeys;
import org.w3c.dom.Node;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.apache.xpath.XPathAPI;

// Imported Serializer classes
import org.apache.xml.serialize.XMLSerializer;

import org.apache.xalan.templates.OutputProperties;

import org.xml.sax.InputSource;
import org.apache.xerces.parsers.DOMParser;
import org.xml.sax.SAXException;
import siena.*;

class Tokenizer {
    private InputStream input;
    private int lnum;
    private int cursor;

    public String sval;

    Tokenizer(InputStream i) throws IOException {
	input = i;
	lnum = 1;
	cursor=input.read();
    }

    public int lineno() {
	return lnum;
    }

    public boolean eof() {
	return cursor == -1;
    }

    public boolean readEquals() throws IOException {
	if (cursor == 0x3d /* '=' */) {
	    cursor = input.read();
	    return true;
	} else {
	    return false;
	}
    }

    public boolean readSemiColon() throws IOException {
	if (cursor == 0x3b /* ';' */) {
	    cursor = input.read();
	    return true;
	} else {
	    return false;
	}
    }

    public void skipWhitesAndComments() throws IOException {
	while(true) {
	    switch(cursor) {
	    case 0x0a: ++lnum;
	    case 0x20:
	    case 0x07:
	    case 0x08:
	    case 0x09:
	    case 0x0b:
	    case 0x0c:
	    case 0x0d:
		cursor = input.read();
		break;
	    case 0x23 /* '#' */: 
		do {
		    cursor = input.read();
		} while (cursor != -1 && cursor != 0x0a /* '\n' */);
		break;
	    default:
		return;
	    }
	}
    }

    private boolean isFirstIdentChar(int c) {
	return (c >= 0x41 && c <= 0x5a)	// 'A' -- 'Z'
	    || (c >= 0x61 && c <= 0x7a)	// 'a' -- 'z'
	    || c == 0x5f;		// '_'
    }

    private boolean isIdentChar(int c) {
	return (c >= 0x41 && c <= 0x5a)	// 'A' -- 'Z' 
	    || (c >= 0x61 && c <= 0x7a)	// 'a' -- 'z'
	    || (c >= 0x30 && c <= 0x39)	// '0' -- '9'
	    || c == 0x5f || c == 0x24	// '_', '$' 
	    || c == 0x2e || c == 0x2f;	// '.', '/' 
    }

    public String readAttributeName() throws IOException {
	if (!isFirstIdentChar(cursor)) return null;
	StringBuffer name = new StringBuffer();
	do {
	    name.append((char)cursor);
	} while(isIdentChar(cursor = input.read()));
	return name.toString();
    }

    public String readId() throws IOException {
	if (!isFirstIdentChar(cursor)) return null;
	StringBuffer name = new StringBuffer();
	do {
	    name.append((char)cursor);
	} while(isFirstIdentChar(cursor = input.read()));
	return name.toString();
    }

    public String readString() throws IOException {
	StringBuffer str = new StringBuffer();
	if (cursor != 0x22 /* '"' */) return null;
	while(true) {
	    cursor = input.read();
	    switch (cursor) {
	    case -1: return null;
	    case 0x22 /* '"' */: 
		cursor = input.read();		
		return str.toString();
	    case 0x5c /* '\\' */: 
		cursor = input.read();
		switch (cursor) {
		case -1: return null;
		case 0x76 /* 'v' */: 
		    str.append((char)0x0b /* '\v' */);break;
		case 0x66 /* 'f' */: 
		    str.append((char)0x0c /* '\f' */);break;
		case 0x72 /* 'r' */: 
		    str.append((char)0x0d /* '\r' */);break; 
		case 0x6e /* 'n' */: 
		    str.append((char)0x0a /* '\n' */); break;
		case 0x74 /* 't' */: 
		    str.append((char)0x09 /* '\t' */); break;
		case 0x62 /* 'b' */: 
		    str.append((char)0x08 /* '\b' */); break;
		case 0x61 /* 'a' */: 
		    str.append((char)0x07 /* '\a' */); break;
		default:
		    //
		    // I should read octal codes, but I don't 
		    // ...work in progress...
		    // 
		    str.append((char)cursor);
		}
		break;
	    default:
		str.append((char)cursor);
	    }
	}
    }
}

/** defines the mapping of XML data into a Siena notification.
 *
 *  Associates attribute names of a Siena notification to rules (see
 *  {@link AttributeRule}).  Once rules are defined for an
 *  <code>EventMap</code>, that <code>EventMap</code> can be used to
 *  process XML documents or in-memory document representations to
 *  produce event notifications.
 **/
public class EventMap {
    private Map	rules;
    private boolean attach_content;

    private static final String XMLContentName = "__xml";

    /** creates an empty map **/
    public EventMap() {
	rules = new HashMap();
	attach_content = false;
    }

    /** instructs this EventMap to make the content of the processed
     *  document available to subscribers.
     *
     *  By default, the XML source document is <i>not</i> transferred.
     **/
    public void attachXml() {
	attach_content = true;
    }

    /** determines whether this EventMap will make the content of the
     *  processed document available to subscribers.
     *
     *  By default, the XML source document is <i>not</i> transferred.
     **/
    public void attachContent(boolean b) {
	attach_content = b;
    }

    private static void putXmlText(Notification n, String xml) {
	String aname = XMLContentName;
	while(n.getAttribute(aname) != null)
	    aname = aname + "_";
	n.putAttribute(aname, xml);
    }

    /** extracts the XML text from a notification produced by an
     *  <code>EventMap</code>
     *
     * @return XML text of the source document that produced this
     *         notification.  Or <code>null</code> is the source
     *         document was not attached to the notification or if
     *         this notification was not generated by an
     *         <code>EventMap</code>
     **/
    public static String getXmlText(Notification n) {
	AttributeValue value = null;
	AttributeValue aux;
	String aname = XMLContentName;

	while((aux = n.getAttribute(aname)) != null) {
	    value = aux;
	    aname = aname + "_";
	}
	if (value == null || value.getType() != AttributeValue.BYTEARRAY) {
	    return null;
	} else {
	    return value.stringValue();
	}
    }

    private static String encode(String s) {
	StringBuffer sb = new StringBuffer();

	byte [] bv = s.getBytes();

	sb.append((char)0x22 /* '"' */);
	for(int i = 0; i < bv.length; ++i)
	    switch(bv[i]) {
	    case 11 /* '\v' */: 
		sb.append((char)0x5c /* '\\' */);
		sb.append((char)0x76 /* 'v' */); 
		break;
	    case 12 /* '\f' */: 
		sb.append((char)0x5c /* '\\' */);
		sb.append((char)0x66 /* 'f' */); 
		break;
	    case 13 /* '\r' */: 
		sb.append((char)0x5c /* '\\' */);
		sb.append((char)0x72 /* 'r' */); 
		break;
	    case 10 /* '\n' */: 
		sb.append((char)0x5c /* '\\' */);
		sb.append((char)0x6e /* 'n' */); 
		break;
	    case 9 /* '\t' */: 
		sb.append((char)0x5c /* '\\' */);
		sb.append((char)0x74 /* 't' */); 
		break;
	    case 8 /* '\b' */: 
		sb.append((char)0x5c /* '\\' */);
		sb.append((char)0x62 /* 'b' */); 
		break;
	    case 7 /* '\a' */: 
		sb.append((char)0x5c /* '\\' */);
		sb.append((char)0x61 /* 'a' */); 
		break;
	    case 0x22 /* '"' */:
		sb.append((char)0x5c /* '\\' */);
		sb.append((char)0x22 /* '"' */); 
		break;
	    case 0x5c /* '\\' */:
		sb.append((char)0x5c /* '\\' */);
		sb.append((char)0x5c /* '\\' */); 
		break;
	    default:
		if (bv[i] < 0x20 || bv[i] >= 0x7F) {
		    //
		    // here I handle other non-printable characters with
		    // the \xxx octal notation ...work in progress...
		    //
		    sb.append((char)0x5c);
		    sb.append(Integer.toOctalString(bv[i]));
		} else {
		    sb.append((char)bv[i]);
		}
	    }
	sb.append((char)0x22 /* '"' */);
	return sb.toString();
    }

    /** saves this map onto an output stream.
     *
     *  The map is saved in a text file, with the syntax specified in
     *  {@link EventMap#importRules(java.io.InputStream)}.
     **/
    public void exportRules(OutputStream os) throws IOException {
	BufferedWriter output = new BufferedWriter(new OutputStreamWriter(os));
	Iterator i = rules.entrySet().iterator();
	while(i.hasNext()) {
	    Map.Entry entry = (Map.Entry)i.next();
	    AttributeRule rule = (AttributeRule)entry.getValue();
	    String name = (String)entry.getKey();
	    switch(rule.type) {
	    default:
	    case AttributeValue.BYTEARRAY:
		output.write("string");
		break;
	    case AttributeValue.LONG:
		output.write("long");
		break;
	    case AttributeValue.BOOL:
		output.write("bool");
		break;
	    case AttributeValue.DOUBLE:
		output.write("double");
		break;
	    }
	    output.write(" " + name + "=" + encode(rule.path));
	    switch(rule.multivalue) {
	    default:
	    case AttributeRule.FIRST_ONLY: output.write(" first"); break;
	    case AttributeRule.LAST_ONLY: output.write(" last"); break;
	    case AttributeRule.XML: output.write(" xml"); break;
	    case AttributeRule.SEPARATOR: 
		output.write(" separator " + encode(rule.separator)); break;
	    }
	    output.write(";\n");
	}
	output.flush();
    }

    /** reads a set of rules from the given input stream.
     *
     *  This method expects a text stream, with the following
     *  (semi-formal) syntax:
     *
     *  <table>
     *  <tr><td><i>&lt;S&gt;</i> 
     *      <td>::= ( <i>&lt;rule&gt;</i> <b><code>;</b> ) +<br>
     *    
     *  <tr><td><i>&lt;rule&gt;</i> 
     *      <td>::= <i>&lt;type&gt;</i> <i>&lt;name&gt;</i> 
     *              <b><code>=</code></b> <i>&lt;path&gt;</i>
     *              [ <i>&lt;multivalue-spec&gt;</i> ]
     *
     *  <tr><td><i>&lt;type&gt;</i>  
     *      <td>::= <b><code>int</code></b> | <b><code>string</code></b> 
     *              | <b><code>double</code></b> | <b><code>bool</code></b><br>
     *
     *  <tr><td><i>&lt;name&gt;</i>  
     *      <td>::= <i>valid Siena attribute name 
     *                 (see Siena.Notification)</i><br>
     *
     *  <tr><td><i>&lt;path&gt;</i>  
     *      <td>::= <i>valid Java/C/C++ string literal</i><br>
     *
     *  <tr><td><i>&lt;multivalue-spec&gt;</i>  
     *      <td>::= <b><code>first</code></b> | <b><code>last</code></b> 
     *              | <b><code>xml</code></b> 
     *              | <b><code>separator</code></b> <i>separator</i><br>
     *  
     *  <tr><td><i>&lt;separator&gt;</i>  
     *      <td>::= <i>valid Java/C/C++ string literal</i><br>
     *
     * </table>
     *
     *  <p>Anything following a pound sign
     *  (<b><code>#</code></b>) up to the first end of line is a
     *  comment.
     *
     *  <p>Example:
     *  <pre><code>
     *      #
     *      #  this is a set of Sxml rules
     *      #
     *      string name="/person/@name";  # default multivalue=first
     *      int age="/person/@age" first;
     *      string parents="/person/parents/person/@name" separator ", ";
     *      string job="/person/job" xml;
     *  </code></pre>
     *
     *  @see EventMap#exportRules(OutputStream)
     **/
    public void importRules(InputStream is) 
	throws RuleParseException, IOException {
	Tokenizer st = new Tokenizer(new BufferedInputStream(is));

	st.skipWhitesAndComments();

	while (!st.eof()) {
	    AttributeRule rule;
	    rule = new AttributeRule();

	    String typename;
	    if ((typename = st.readId()) == null)
 		throw new RuleParseException(st.lineno() + ": expecting type name");

	    if (typename.equals("string") || typename.equals("bytearray")) {
		rule.type = AttributeValue.BYTEARRAY;
	    } else if (typename.equals("int") || typename.equals("long")) {
		rule.type = AttributeValue.LONG;
	    } else if (typename.equals("double") || typename.equals("float")) {
		rule.type = AttributeValue.DOUBLE;
	    } else if (typename.equals("bool")) {
		rule.type = AttributeValue.BOOL;
	    } else {
 		throw new RuleParseException(st.lineno() 
					     + ": unknown type name `" 
					     + typename + "'");
	    }

	    st.skipWhitesAndComments();

	    String name;

	    if ((name = st.readAttributeName()) == null)
 		throw new RuleParseException(st.lineno() 
					     + ": expecting attribute name");

	    st.skipWhitesAndComments();

	    if (!st.readEquals())
		throw new RuleParseException(st.lineno()
					     + ": expecting `='");

	    st.skipWhitesAndComments();

	    if ((rule.path = st.readString()) == null)
		throw new RuleParseException(st.lineno()
					     + ": expecting path");
	    st.skipWhitesAndComments();

	    if (!st.readSemiColon()) {
		String qualif;
		if ((qualif = st.readId()) == null) 
		    throw new RuleParseException(st.lineno() + ": expecting `;' or multivalue qualifier"); 		    
		if (qualif.equals("first")) {
		    rule.multivalue = AttributeRule.FIRST_ONLY;
		} else if (qualif.equals("last")) {
		    rule.multivalue = AttributeRule.LAST_ONLY;
		} else if (qualif.equals("xml")) {
		    rule.multivalue = AttributeRule.XML;
		} else if (qualif.equals("separator")) {
		    rule.multivalue = AttributeRule.SEPARATOR;
		    st.skipWhitesAndComments();
		    if ((rule.separator = st.readString()) == null) 
			throw new RuleParseException(st.lineno() + 
						     ": expecting separator"); 
		}
		st.skipWhitesAndComments();
		if (!st.readSemiColon())
		    throw new RuleParseException(st.lineno() 
						 + ": expecting `;'");
	    }
	    putRule(name, rule);
	    st.skipWhitesAndComments();
	}
    }

    /** reads a set of rules from the given input file.
     *  
     *  @see EventMap#importRules(java.io.InputStream)
     **/
    public void importRules(String filename) 
    throws RuleParseException, IOException {
	importRules(new FileInputStream(filename));
    }

    /** adds or overwrites a mapping rule for the given attribute name 
     **/
    public void putRule(String attr_name, AttributeRule rule) {
	rules.put(attr_name, rule);
    }

    /** removes the mapping rule for the given attribute name 
     **/
    public AttributeRule removeRule(String attr_name) {
	return (AttributeRule)rules.remove(attr_name);
    }

    /** returns the mapping rule associated with the given attribute name 
     **/
    public AttributeRule getRule(String attr_name) {
	return (AttributeRule)rules.get(attr_name);
    }

    /** processes an XML document to produce a Siena notification **/ 
    public Notification makeNotification(InputStream is) 
	throws SAXException, IOException, TransformerException,
	       EventMapException {
	InputSource input;
	input = new InputSource(is);

	DOMParser parser = new DOMParser();
	parser.parse(input);
	return makeNotification(parser.getDocument().getDocumentElement());
    }

    /** processes an XML document to produce a Siena notification **/ 
    public Notification makeNotification(String filename) 
	throws SAXException, IOException, TransformerException, 
	       EventMapException {
	return makeNotification(new FileInputStream(filename));
    }

    /** processes an XML document to produce a Siena notification **/ 
    public Notification makeNotification(Node root) 
	throws TransformerException, EventMapException {
	Iterator ri = rules.entrySet().iterator();
	Notification event = new Notification();
	while(ri.hasNext()) {
	    Map.Entry entry = (Map.Entry)ri.next();
	    String name = (String)entry.getKey();
	    AttributeRule rule = (AttributeRule)entry.getValue();
	    NodeList nl = XPathAPI.selectNodeList(root, rule.path);
	    String value;
	    value = null;
	    if (nl.getLength() > 0) {
		switch (rule.multivalue) {
		case AttributeRule.SEPARATOR:
		    StringBuffer sbuf = null;
		    for(int i = 0; i < nl.getLength(); ++i) {
			Node n = nl.item(i);
			if (n.getNodeType() == Node.CDATA_SECTION_NODE
			    || n.getNodeType() == Node.ATTRIBUTE_NODE) {
			    if (sbuf == null) {
				sbuf = new StringBuffer();
				sbuf.append(rule.separator);
			    }
			    sbuf.append(n.getNodeValue());
			    sbuf.append(rule.separator);
			}
		    }
		    value = (sbuf != null) ? sbuf.toString() : null;
		    break;
		case AttributeRule.XML:
		    ByteArrayOutputStream out = null;
		    try {
			XMLSerializer sxml = null;
			for(int i = 0; i < nl.getLength(); ++i) {
			    Node n = nl.item(i);
			    if (n.getNodeType() == Node.ELEMENT_NODE) {
				if (sxml == null) {
				    out = new ByteArrayOutputStream();
				    sxml = new XMLSerializer();
				    sxml.setOutputByteStream(out);
				}
				sxml.serialize((Element)n);
			    }
			}
		    } catch (IOException ex) {
			//
			// I don't know what to do with this
			// exception.  It should never be thrown
			// anyway since I'm using a ByteArrayOutputStream.
			// ...work in progress...
			//
			ex.printStackTrace();
		    }
		    value = (out != null) ? out.toString() : null;
		    break;
		case AttributeRule.LAST_ONLY:
		    value = nl.item(nl.getLength()-1).getNodeValue();
		    break;
		case AttributeRule.FIRST_ONLY:
		default:
		    value = nl.item(0).getNodeValue();
		    break;
		}
		if (value != null) {
		    try {
			switch(rule.type) {
			case AttributeValue.BYTEARRAY:
			    event.putAttribute(name, value);
			    break;
			case AttributeValue.LONG:
			    event.putAttribute(name, Long.parseLong(value));
			    break;
			case AttributeValue.DOUBLE:
			    event.putAttribute(name, Double.parseDouble(value));
			    break;
			case AttributeValue.BOOL:
			    event.putAttribute(name, Boolean.valueOf(value).booleanValue());
			    break;		
			}
		    } catch (NumberFormatException ex) {
			throw(new EventMapException(ex.toString()));
		    }
		}
	    }
	}
	if (attach_content) {
	    ByteArrayOutputStream out;
	    out = new ByteArrayOutputStream();
	    try {
		XMLSerializer sxml = new XMLSerializer();

		sxml.setOutputByteStream(out);
		sxml.serialize(root.getOwnerDocument());
		out.flush();
		putXmlText(event, out.toString());
	    } catch (IOException ex) {
		//
		// I don't know what to do with this
		// exception.  It should never be thrown
		// anyway since I'm using a ByteArrayOutputStream.
		// ...work in progress...
		//
		ex.printStackTrace();
	    }
	}

	return event;
    }
}
