//
//  This file is part of SXML, a tool to automate the publication of
//  XML data through the Siena wide-area event notification system.
//
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Copyright (C) 2000-2001 University of Colorado
//
//  Author: Antonio Carzaniga <carzanig@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
// $Id: XMLNotifiableWrapper.java,v 1.3 2001/01/24 17:32:14 carzanig Exp $
//
package siena.sxml;

import siena.Notification;
import siena.Notifiable;

/** a wrapper for using an <code>XMLNotifiable</code> as a subscriber with
 *  Siena.
 * 
 *  serves as an adapter for an {@link XMLNotifiable} to be used as a
 *  Siena notifiable. 
 *
 *  <p>Example:
 *  <pre><code>
 *      Siena siena = new ThinClient("senp://foo.bar:4321");
 *      XMLNotifiable subscriber;
 *      //
 *      //  ...initialize subscriber...
 *      //
 *      XMLNotifiableWrapper wrapper = new XMLNotifiableWrapper(subscriber);
 *
 *      Filter f = new Filter();
 *      f.addConstraint("name", Op.EQ, "Antonio");
 *
 *      siena.subscribe(wrapper, f);
 *      //
 *      // ...
 *      //
 *      siena.unsubscribe(wrapper);
 *  </code></pre>
 *  
 *  an <code>XMLNotifiableWrapper</code> extracts the XML fragments
 *  from the notifications it receives from Siena and dispatches them
 *  to the XMLNotifiable.  If a notification does not contain any XML
 *  fragment, because it was not generated by {@link EventMap} or
 *  because it was produced without the XML source (see {@link
 *  EventMap#attachXml()}), then <code>XMLNotifiableWrapper</code>
 *  will pass a <code>null</code> string.
 **/
public class XMLNotifiableWrapper implements Notifiable {
    private XMLNotifiable notifiable;
    
    /** construct a wrapper for the given <code>XMLNotifiable</code>
     **/
    public XMLNotifiableWrapper(XMLNotifiable n) {
	notifiable = n;
    }

    public void notify(Notification n) {
	notifiable.notifyXml(EventMap.getXmlText(n));
    }

    public void notify(Notification [] ns) {
	String [] events = new String[ns.length];
	for(int i = 0; i < ns.length; ++i)
	    events[i] = EventMap.getXmlText(ns[i]);

	notifiable.notifyXml(events);
    }
}
